'''module for dealing with Moab's checknode output'''

from xml.dom import minidom

class ChecknodeParser(object):
    '''Parser class for Moab checknode ouptut'''

    def parse(self, checknode_xml):
        '''parse checknode XML output, and return the list of features'''
        dom = minidom.parseString(checknode_xml)
        node = dom.getElementsByTagName('node')[0]
        feature_str = node.getAttribute('FEATURES')
        return feature_str.split(',')

    def parse_file(self, node_file):
        '''parse a file that contains checknode output'''
        node_output = ''.join(node_file.readlines())
        return self.parse(node_output)



#!/usr/bin/env python

import re
import sqlite3
from subprocess import check_output, CalledProcessError, STDOUT
from vsc.utils import bytes2size
from vsc.event_logger import EventLogger
from vsc.mam.gbalance import GbalanceParser
from vsc.mam.quote import QuoteCalculator


class JobChecker(EventLogger):
    '''Semantic checker for jobs'''

    def __init__(self, config, event_defs):
        '''Constructor for job checker'''
        super(JobChecker, self).__init__(event_defs, context='global')
        self._conn = sqlite3.connect(config['cluster_db'])
        self._cursor = self._conn.cursor()
        self._config = config

    def check(self, job):
        '''Check semantics of given job'''
        self.check_partition(job)
        self.check_properties(job)
        self.check_features(job)
        self.check_ppn(job)
        self.check_qos(job)
        self.check_total_pmem(job)
        self.check_mem(job)
        self.check_mem_vs_pmem(job)
        if self._config['check_accounting']:
            self.check_credit_account(job)

    def check_qos(self, job):
        '''check QOS specified exists'''
        job_qos = job.resource_spec('qos')
        if job_qos is not None:
            qos = self._qos()
            if job_qos not in qos:
                self.reg_event('unknown_qos',
                               {'qos': job_qos})

    def check_properties(self, job):
        '''check node properties specified exists'''
        node_specs = job.resource_spec('nodes')
        properties = self._properties()
        for node_spec in node_specs:
            if 'properties' in node_spec:
                for property in node_spec['properties']:
                    if property not in properties:
                        self.reg_event('unknown_property',
                                       {'property': property})

    def check_features(self, job):
        '''check features specified exists'''
        features = self._features()
        if job.resource_spec('feature'):
            job_features = job.resource_spec('feature')
            for feature in job_features:
                if feature not in features:
                    self.reg_event('unknown_feature',
                                   {'feature': feature})

    def check_partition(self, job):
        '''check whether the specified partition exists, and whether
           it has the total number of requested nodes'''
        partition = job.resource_spec('partition')
        partitions = self._partitions()
        if partition not in partitions:
            self.reg_event('unknown_partition',
                           {'partition': partition})
        else:
            nodes = 0
            for nodes_spec in job.resource_spec('nodes'):
                nodes += nodes_spec['nodes']
            if nodes > partitions[partition]:
                self.reg_event('insufficient_nodes',
                               {'nodes': nodes,
                                'max_nodes': partitions[partition]})

    def check_ppn(self, job):
        '''check whether there are enough nodes with the requested ppn'''
        partition = job.resource_spec('partition')
        all_ppn = self._ppn(partition)
        for node_spec in job.resource_spec('nodes'):
            if 'ppn' in node_spec:
                job_ppn = node_spec['ppn']
                job_nodes = node_spec['nodes']
                for ppn in sorted(all_ppn):
                    if ppn >= job_ppn:
                        if job_nodes <= all_ppn[ppn]:
                            all_ppn[ppn] -= job_nodes
                            job_nodes = 0
                            break
                        else:
                            job_nodes -= all_ppn[ppn]
                            all_ppn[ppn] = 0
                if job_nodes > 0:
                    self.reg_event('insufficient_ppn_nodes',
                                   {'ppn': job_ppn})

    def check_pmem(self, job):
        '''Check whether the requested memory per node is available'''
        partition = job.resource_spec('partition')
        mem_sizes = list(self._mem_sizes(partition).keys())
        for nodes_spec in job.resource_spec('nodes'):
            satisfied = False
            ppn = nodes_spec['ppn']
            pmem = job.resource_spec('pmem')
            if ppn and pmem:
                node_mem = ppn*pmem
            else:
                continue
            for mem in mem_sizes:
                if node_mem < mem:
                    satisfied = True
                    break
            if not satisfied:
                self.reg_event('insufficient_node_pmem',
                               {'mem': bytes2size(node_mem, 'gb')})

    def check_total_pmem(self, job):
        '''check total memory requirements of job as aggregated of pmem'''
        partition = job.resource_spec('partition')
        mem_sizes = self._mem_sizes(partition)
        for nodes_spec in job.resource_spec('nodes'):
            orig_nodes = nodes_spec['nodes']
            nodes = orig_nodes
            if 'ppn' in nodes_spec:
                ppn = nodes_spec['ppn']
            else:
                ppn = None
            pmem = job.resource_spec('pmem')
            if ppn and pmem:
                node_mem = ppn*pmem
            else:
                continue
            for mem in sorted(mem_sizes.keys()):
                if node_mem < mem:
                    if nodes <= mem_sizes[mem]:
                        mem_sizes[mem] -= nodes
                        nodes = 0
                        break
                    else:
                        nodes -= mem_sizes[mem]
            if nodes > 0:
                self.reg_event('insufficient_nodes_mem',
                               {'mem': bytes2size(node_mem, 'gb'),
                                'nodes': orig_nodes})

    def check_mem(self, job):
        '''check total memory requirements of job as mem'''
        partition = job.resource_spec('partition')
        mem_sizes = self._mem_sizes(partition)
        mem_spec = job.resource_spec('mem')
        if mem_spec:
            for nodes_spec in job.resource_spec('nodes'):
                nodes = nodes_spec['nodes']
                for mem in sorted(mem_sizes.keys()):
                    mem_spec -= mem*nodes
                    if mem_spec <= 0:
                        break
            if mem_spec > 0:
                self.reg_event('insufficient_mem',
                               {'mem': bytes2size(mem_spec, 'gb')})

    def check_mem_vs_pmem(self, job):
        '''Check whether both mem and pmem are specified'''
        if job.resource_spec('mem') and not job.has_default_pmem:
            self.reg_event('both_mem_pmem_specs')

    def check_credit_account(self, job):
        '''Check whether a project is available to debit credits from,
           and whether the balance is sufficient'''
        if 'mock_balance' in self._config:
            with open(self._config['mock_balance'], 'r') as balance_file:
                balance_sheet = ''.join(balance_file.readlines())
        else:
            balance_cmd = self._config['balance_cmd']
            try:
                balance_sheet = check_output([balance_cmd], stderr=STDOUT)
            except OSError:
                return
            except CalledProcessError:
                # TODO: decide on user feedback
                return
        accounts = GbalanceParser().parse(balance_sheet)
        if len(accounts) == 0:
            self.reg_event('no_credit_account')
            return
        elif job.project is None:
            credit_account = None
            for account_id in accounts:
                account_name = accounts[account_id].name
                if (account_name == '' or
                        account_name == self._config['default_project']):
                    credit_account = accounts[account_id]
                    break
            if not credit_account:
                self.reg_event('no_default_credit_account')
                return
        else:
            credit_account = None
            for account_id in accounts:
                if accounts[account_id].name == job.project:
                    credit_account = accounts[account_id]
                    break
            if not credit_account:
                self.reg_event('unknown_credit_account',
                               {'account': job.project})
                return
        quoteCalculator = QuoteCalculator(self._config)
        credits = quoteCalculator.compute(job)
        if credits > credit_account.available_credits:
            self.reg_event('insufficient_credits',
                           {'account': credit_account.name})

    def _mem_sizes(self, partition):
        '''retrieve the memory sizes of the nodes from the databse'''
        sizes = {}
        stmt = '''SELECT n.mem, count(n.node_id)
                      FROM nodes as n NATURAL JOIN partitions as p
                      WHERE p.partition_name = ?
                      GROUP BY mem'''
        self._cursor.execute(stmt, (partition, ))
        for row in self._cursor:
            sizes[row[0]] = row[1]
        return sizes

    def _partitions(self):
        '''retrieve the list of partitions and their nodes from databasse'''
        partitions = {}
        stmt = '''SELECT p.partition_name, count(n.node_id)
                      FROM partitions as p NATURAL JOIN nodes as n
                      GROUP BY p.partition_name'''
        self._cursor.execute(stmt)
        for row in self._cursor:
            partitions[row[0]] = row[1]
        return partitions

    def _qos(self):
        '''retrieve list of QOS levels'''
        qos = []
        stmt = '''SELECT qos FROM qos_levels'''
        self._cursor.execute(stmt)
        for row in self._cursor:
            qos.append(row[0])
        return qos

    def _features(self):
        '''retrieve list of features'''
        features = []
        stmt = '''SELECT DISTINCT feature FROM features'''
        self._cursor.execute(stmt)
        for row in self._cursor:
            features.append(row[0])
        return features

    def _properties(self):
        '''retrieve list of propertie and propertiess'''
        properties = []
        stmt = '''SELECT DISTINCT property FROM properties'''
        self._cursor.execute(stmt)
        for row in self._cursor:
            properties.append(row[0])
        return properties

    def _ppn(self, partition):
        '''retrieve the number of nodes per ppn'''
        ppn = {}
        stmt = '''SELECT n.np, count(*)
                      FROM nodes as n NATURAL JOIN partitions as p
                      WHERE p.partition_name = ?
                      GROUP BY n.np'''
        self._cursor.execute(stmt, (partition, ))
        for row in self._cursor:
            ppn[row[0]] = row[1]
        return ppn


class ScriptChecker(EventLogger):
    '''Implementation of a bash script checker that uses some heuristics
       to check for common errors'''

    def __init__(self, config, event_defs):
        '''Constructor for script checker'''
        super(ScriptChecker, self).__init__(event_defs)
        try:
            from fuzzywuzzy import fuzz
            self._fuzz = fuzz
        except ImportError:
            self._fuzz = None

    def check(self, job, start_line_nr):
        '''check a script for potential errors'''
        self._line_nr = start_line_nr - 1
        for line in job.script.split('\n'):
            self._line_nr += 1
            self._check_module_load(line)
            self._check_workdir(line)

    def _check_module_load(self, line):
        '''check whether this line contains a module load, if so do
           some basic validity checks'''
        if not self._fuzz:
            return
        line = line.strip()
        cmd = 'module load'
        cmd_re = r'module\s+load\s+.*'
        ratio = self._fuzz.partial_token_sort_ratio(cmd, line)
        if ratio > 75:
            if not re.match(cmd_re, line):
                self.reg_event('missspelled', {'correct': cmd})

    def _check_workdir(self, line):
        '''check whether this line contains PBS_O_WORKDIR and performs
           some basic validity checks'''
        if not self._fuzz:
            return
        line = line.strip()
        var = 'PBS_O_WORKDIR'
        ratio = self._fuzz.partial_token_sort_ratio(var.lower(),
                                                    line.lower())
        if ratio > 75:
            if var not in line:
                self.reg_event('missspelled',
                               {'correct': '${{{0}}}'.format(var)})
            else:
                var_value = '${{{0}}}'.format(var)
                if not('${0}'.format(var) in line or
                       var_value in line):
                    self.reg_event('missspelled', {'correct': var_value})



#!/usr/bin/env python
'''Utilities to deal with PBS torque qstat'''

from vsc.utils import walltime2seconds
from vsc.pbs.job import PbsJob

class QstatParser(object):
    '''Parser for full PBS torque qstat output'''

    def __init__(self, config):
        '''constructor'''
        self._config = config
        self._jobs = {}

    def _get_value(self, line):
        '''extract value from line'''
        _, value = line.split('=', 1)
        return value.strip()

    def parse_file(self, qstat_file):
        '''parse a file that contains qstat -f output'''
        qstat_output = ''.join(qstat_file.readlines())
        return self.parse(qstat_output)

    def parse_record(self, record):
        '''parse an individual job record'''
        job = None
        resource_specs = {}
        resources_used = {}
        state = None
        host_str = None
        for line in record.split('\n'):
            line = line.strip()
            if state == 'exec_host':
                if not line.startswith('exec_port'):
                    host_str += line
                    continue
                else:
                    hosts = {}
                    for host in host_str.split('+'):
                        node, core = host.split('/')
                        if node not in hosts:
                            hosts[node] = []
                        hosts[node].append(core)
                    job.exec_host = hosts
                    state = None
                    host_str = None
            if line.startswith('Job Id:'):
                _, job_id = line.split(':', 1)
                job = PbsJob(self._config, job_id.strip())
            elif line.startswith('Job_Name ='):
                job.name = self._get_value(line)
            elif line.startswith('euser ='):
                job.user = self._get_value(line)
            elif line.startswith('job_state = '):
                job.state = self._get_value(line)
            elif line.startswith('queue ='):
                job.queue = self._get_value(line)
            elif line.startswith('Account_Name ='):
                job.project = self._get_value(line)
            elif line.startswith('resources_used.walltime ='):
                walltime = self._get_value(line)
                resources_used['walltime'] = walltime2seconds(walltime)
            elif line.startswith('Resource_List.walltime ='):
                walltime = self._get_value(line)
                resource_specs['walltime'] = walltime2seconds(walltime)
            elif line.startswith('Resource_List.nodect = '):
                nodect = int(self._get_value(line))
                resource_specs['nodect'] = nodect
            elif line.startswith('exec_host ='):
                host_strs = self._get_value(line) .split('+')
                exec_host = dict()
                for host_str in host_strs:
                    if '/' in host_str:
                        host, cores = host_str.split('/')
                        exec_host[host] = cores
                    else:
                        exec_host[host_str] = None
                job.exec_host = exec_host
            elif line.startswith('Resource_List.partition ='):
                job.partition = self._get_value(line)
            elif line.startswith('qtime = '):
                job.queue_time = self._get_value(line)
            elif line.startswith('start_time = '):
                job.start_time = self._get_value(line)
        job.add_resource_specs(resource_specs)
        job.add_resources_used(resources_used)
        return job

    def parse(self, qstat_str):
        '''parse PBS torque qstat full output, and return list of jobs'''
        jobs = []
        job_str = None
        for line in qstat_str.split('\n'):
            if line.startswith('Job Id:'):
                if job_str:
                    jobs.append(self.parse_record(job_str))
                job_str = line
            elif line.strip():
                job_str += '\n' + line
        if job_str:
            jobs.append(self.parse_record(job_str))
        return jobs




#!/usr/bin/env python
'''Utilities to parse optput of Adaptive Moab showq output'''

import re

from vsc.moab.job import JobStatus

class ShowqParser(object):
    '''Parser for Moab showq output'''

    def __init__(self):
        '''Moab showq output parser constructor'''
        pass

    def parse(self, showq_output):
        '''Returns map of jobs for given showq output'''
        jobs = {
            'active': [],
            'eligible': [],
            'blocked': []
        }
        parser_state = 'init'
        for line in showq_output.split('\n'):
            if len(line) == 0:
                pass
            elif line.startswith('active'):
                parser_state = 'active'
            elif line.startswith('eligible'):
                parser_state = 'eligible'
            elif line.startswith('blocked'):
                parser_state = 'blocked'
            elif re.match(r'\d+\s+(active|eligible|blocked)', line):
                pass
            elif line[0].isdigit():
                line = re.sub(r'\s+', ' ', line)
                jobid, username, state, procs, time, date = line.split(' ', 5)
                job = JobStatus(id=jobid, username=username, state=state,
                                procs=procs, time=time, date=date)
                jobs[parser_state].append(job)
        return jobs

    def parse_file(self, showq_file):
        '''Returns map of jobs for given showq output as file'''
        showq_output = ''.join(showq_file.readlines())
        return self.parse(showq_output)




#!/usr/bin/env python
'''class for parsing PBS options'''

from argparse import ArgumentParser
import os
import re
import validate_email

from vsc.event_logger import EventLogger
from vsc.utils import walltime2seconds, size2bytes
from vsc.utils import InvalidWalltimeError


class PbsOptionParser(EventLogger):
    '''Parser for PBS options, either command line or directives'''

    def __init__(self, config, event_defs, job):
        '''constructor'''
        super(PbsOptionParser, self).__init__(event_defs, 'global')
        self._config = config
        self._job = job
        self._arg_parser = ArgumentParser()
        self._arg_parser.add_argument('-a')
        self._arg_parser.add_argument('-A')
        self._arg_parser.add_argument('-e')
        self._arg_parser.add_argument('-j')
        self._arg_parser.add_argument('-k')
        self._arg_parser.add_argument('-l', action='append')
        self._arg_parser.add_argument('-m')
        self._arg_parser.add_argument('-M')
        self._arg_parser.add_argument('-N')
        self._arg_parser.add_argument('-o')
        self._arg_parser.add_argument('-q')

    def parse_args(self, option_line):
        '''parse options string'''
        self._events = []
        args = option_line.split()
        options, rest = self._arg_parser.parse_known_args(args)
        for option, value in options.__dict__.items():
            if value:
                self.handle_option(option, value)
        if self._job.queue and not self._job._is_time_limit_set:
            walltime_limit = self.get_queue_limit(self._job.queue)
            if walltime_limit:
                self._job._resource_specs['walltime'] = walltime_limit

    def handle_option(self, option, value):
        '''option dispatch method'''
        if option == 'a':
            self.check_a(value.strip())
        elif option == 'A':
            self.check_A(value.strip())
        elif option == 'e' or option == 'o':
            self.check_oe(value.strip(), option)
        elif option == 'j':
            self.check_j(value.strip())
        elif option == 'k':
            self.check_k(value.strip())
        elif option == 'l':
            self.check_l(value)
        elif option == 'm':
            self.check_m(value.strip())
        elif option == 'M':
            self.check_M(value.strip())
        elif option == 'N':
            self.check_N(value.strip())
        elif option == 'q':
            self.check_q(value.strip())

    def is_valid_datetime(self, dt_str):
        regex = r'^\s*(?:(?:(?:(?:(?:\d{2})?\d{2})?\d{2})?\d{2})?\d{2})?\d{4}(?:\.\d{2})?\s*$'
        match = re.match(regex, dt_str)
        return match is not None

    def check_a(self, val):
        '''check whether a valid datetime string was specified'''
        if self.is_valid_datetime(val):
            pass
        else:
            self.reg_event('invalid_datetime', {'val': val})

    def check_A(self, val):
        '''check whether a valid project name was specified'''
        if re.match(r'[A-Za-z]\w*$', val):
            self._job.project = val
        else:
            self.reg_event('invalid_project_name', {'val': val})

    def get_queue_limit(self, queue_name):
        '''get the maximum walltime for the queue specified'''
        for queue_def in self._config['queue_definitions']:
            if queue_def['name'] == queue_name:
                return int(queue_def['walltime_limit'])
        return None

    def check_q(self, val):
        '''check whether a valid queue name was specified'''
        if re.match(r'[A-Za-z]\w*$', val):
            self._job.queue = val
        else:
            self.reg_event('invalid_queue_name', {'val': val})

    def check_j(self, val):
        '''check -j option, vals can be oe, eo, n'''
        if val == 'oe' or val == 'eo' or val == 'n':
            self._job.join = val
        else:
            self.reg_event('invalid_join', {'val': val})

    def check_k(self, val):
        '''check -k option, val can be e, o, oe, eo, or n'''
        if re.match(r'^[eo]+$', val) or val == 'n':
            self._job.keep = val
        else:
            self.reg_event('invalid_keep', {'val': val})

    def check_m(self, val):
        '''check -m option, val can be any combination of b, e, a, or n'''
        if re.match(r'^[bea]+$', val) or val == 'n':
            self._job.mail_events = val
        else:
            self.reg_event('invalid_mail_event', {'val': val})

    def check_M(self, val):
        '''check -M option'''
        self._job.mail_addresses = val.split(',')
        uid = os.getlogin()
        for address in self._job.mail_addresses:
            if (not validate_email.validate_email(address) and
                    address != uid):
                self.reg_event('invalid_mail_address', {'address': address})

    def check_N(self, val):
        '''check -N is a valid job name'''
        if re.match(r'[A-Za-z]\w{,14}$', val):
            self._job.name = val
        else:
            self.reg_event('invalid_job_name', {'val': val})

    def check_time_res(self, val, resource_spec):
        '''check a time resource'''
        attr_name, attr_value = val.split('=')
        try:
            seconds = walltime2seconds(attr_value)
            resource_spec[attr_name] = seconds
        except InvalidWalltimeError:
            self.reg_event('invalid_{0}_format'.format(attr_name),
                           {'time': attr_value})

    def check_generic_res(self, val, resource_spec):
        '''check a generic resource'''
        attr_name, attr_value = val.split('=')
        if attr_name == 'feature':
            resource_spec[attr_name] = attr_value.split(':')
        else:
            resource_spec[attr_name] = attr_value

    def check_mem_res(self, val, resource_spec):
        '''check memory resource'''
        attr_name, attr_value = val.split('=')
        match = re.match(r'(\d+)([kmgt])?[bw]', attr_value)
        if match:
            amount = int(match.group(1))
            order = match.group(2)
            resource_spec[attr_name] = size2bytes(amount, order)
        else:
            self.reg_event('invalid_{0}_format'.format(attr_name),
                           {'size': attr_value})

    def check_nodes_res(self, val, resource_spec):
        '''check nodes resource'''
        _, attr_value = val.split('=', 1)
        node_specs = PbsOptionParser.parse_node_spec_str(attr_value, self)
        resource_spec['nodes'] = node_specs

    def check_procs_res(self, val, resource_spec):
        '''check procs resource specification'''
        attr_name, attr_value = val.split('=')
        if attr_name in resource_spec:
            self.reg_event('multiple_procs_specs')
        if not attr_value.isdigit():
            self.reg_event('non_integer_procs', {'procs': attr_value})
        resource_spec[attr_name] = int(attr_value)

    def check_l(self, vals):
        '''check and handle resource options'''
        resource_spec = {}
        has_default_pmem = True
        # there can be multiple -l options on one line or on command line
        for val_str in (x.strip() for x in vals):
            # values can be combined by using ','
            for val in (x.strip() for x in val_str.split(',')):
                if (val.startswith('walltime=') or
                        val.startswith('cput=') or
                        val.startswith('pcput=')):
                    self.check_time_res(val, resource_spec)
                    self._job._is_time_limit_set = True
                elif (val.startswith('mem=') or val.startswith('pmem=') or
                      val.startswith('vmem=') or val.startswith('pvmem=')):
                    self.check_mem_res(val, resource_spec)
                    if val.startswith('pmem='):
                        has_default_pmem = False
                elif val.startswith('nodes='):
                    self.check_nodes_res(val, resource_spec)
                elif val.startswith('procs='):
                    self.check_procs_res(val, resource_spec)
                elif (val.startswith('partition=') or
                      val.startswith('feature') or
                      val.startswith('qos')):
                    self.check_generic_res(val, resource_spec)
                else:
                    self.reg_event('unknown_resource_spec', {'spec': val})
        self._job.add_resource_specs(resource_spec)
        self._job._has_default_pmem = has_default_pmem

    def check_oe(self, val, option):
        '''check for valid -o or -e paths'''
        if ':' in val:
            host, path = val.split(':', 1)
        else:
            host = None
            path = val
        if option == 'e':
            self._job.set_error(path, host)
        else:
            self._job.set_output(path, host)

    @staticmethod
    def parse_node_spec_str(attr_value, parser=None):
# if present, multiple node specifications are separated by '+'
        node_spec_strs = attr_value.split('+')
        node_specs = []
        for node_spec_str in node_spec_strs:
            node_spec = {'properties': []}
            spec_strs = node_spec_str.split(':')
# if a node spec starts with a number, that's the number of nodes,
# otherwise it can be a hostname or a feature, but number of nodes is 1
            if spec_strs[0].isdigit():
                node_spec['nodes'] = int(spec_strs[0])
            else:
                node_spec['nodes'] = 1
# note that this might be wrong, it may actually be a feature, but
# that is a semantic check, not syntax
                node_spec['host'] = spec_strs[0]
# now deal with the remaining specifications, ppn, gpus and properties
            for spec_str in spec_strs[1:]:
                if (spec_str.startswith('ppn=') or
                        spec_str.startswith('gpus=')):
                    key, value = spec_str.split('=')
                    if value.isdigit():
                        node_spec[key] = int(value)
                    elif parser:
                        parser.reg_event('{0}_no_number'.format(key),
                                         {'number': value})
                else:
                    node_spec['properties'].append(spec_str)
            node_specs.append(node_spec)
        return node_specs



'''Module to handle PBS job events'''

from datetime import datetime
import re

from vsc.pbs.option_parser import PbsOptionParser
from vsc.utils import walltime2seconds, size2bytes


class PbsJobEvent(object):
    '''Implements job events'''

    _int_keys = [
        'qtime', 'ctime', 'etime', 'start', 'end',
        'Resource_List.nodect',
        'total_execution_slots', 'unique_node_count', 'Exit_status',
        'resources_used.cput', 'resources_used.walltime',
    ]
    _time_keys = [
        'Resource_List.walltime',
    ]
    _mem_keys = [
        'Resource_List.mem', 'Resource_List.pmem', 'Resource_List.vmem',
        'resources_used.mem', 'resources_used.vmem',
    ]

    _key_map = {
        'jobname': 'name',
        'user': 'user',
        'Resource_List.partition': 'partition',
        'account': 'project',
        'queue': 'queue',
        'Exit_status': 'exit_status',
        'exec_host': 'exec_host',
    }

    def __init__(self, time_stamp, event_type, info_str):
        '''Constructor'''
        self._time_stamp = datetime.strptime(time_stamp,
                                             '%m/%d/%Y %H:%M:%S')
        self._type = event_type
        self._info = PbsJobEvent._parse_info(info_str)

    @property
    def time_stamp(self):
        '''returns time stamp for the event as a datetime object'''
        return self._time_stamp

    @property
    def type(self):
        '''returns event type'''
        return self._type

    def is_queue(self):
        '''returns True if this is a queue event'''
        return self.type == 'Q'

    def is_start(self):
        '''returns True if this is a start event'''
        return self.type == 'S'

    def is_end(self):
        '''returns True if this is a end event'''
        return self.type == 'E'

    def is_delete(self):
        '''returns True if this is a delete event'''
        return self.type == 'D'

    def has_info(self, key):
        '''return True if the event information has the specified key'''
        return key in self._info

    def info(self, key):
        '''return value in the event information for the specified key'''
        if self.has_info(key):
            return self._info[key]
        else:
            return None

    def __str__(self):
        '''returns string representation of the event'''
        time_stamp = self.time_stamp.strftime('%Y-%m-%d %H:%M:%S')
        info_strs = ['  {0}: {1}'.format(key, self._info[key])
                     for key in self._info]
        return '{0}: {1}\n{2}'.format(self.type, time_stamp,
                                      '\n'.join(info_strs))

    def update_job_info(self, job):
        '''update the job info based on this event'''
        job.state = self.type
        for key, attr_name in PbsJobEvent._key_map.items():
            if self.has_info(key):
                job.__setattr__(attr_name, self.info(key))
        prefix = 'Resource_List.'
        for key in self._info:
            if key.startswith(prefix):
                resource = key[len(prefix):]
                if resource == 'nodes':
                    node_specs = PbsOptionParser.parse_node_spec_str(self.info(key))
                    job.add_resource_spec(resource, node_specs)
                elif resource == 'feature':
                    job.add_resource_spec('features', [self.info(key)])
                else:
                    job.add_resource_spec(resource, self.info(key))
        prefix = 'resources_used.'
        for key in self._info:
            if key.startswith(prefix):
                resource = key[len(prefix):]
                job.add_resource_used(resource, self.info(key))

    @staticmethod
    def _parse_info(info_str):
        '''parse a job event info string into a dictionary'''
        key_value_strs = re.split(r'\s+', info_str)
        info = dict()
        for key_value_str in key_value_strs:
            if key_value_str:
                key, value = key_value_str.split('=', 1)
                info[key] = value
        PbsJobEvent._format_info(info)
        return info

    @staticmethod
    def _format_info(info):
        '''format values in info dictionary'''
        for key in PbsJobEvent._int_keys:
            if key in info:
                info[key] = int(info[key])
        for key in PbsJobEvent._time_keys:
            if key in info:
                info[key] = walltime2seconds(info[key])
        for key in PbsJobEvent._mem_keys:
            if key in info:
                info[key] = size2bytes(info[key])
        key = 'exec_host'
        if key in info:
            exec_host = dict()
            for exec_host_str in info[key].split('+'):
                host, cores = exec_host_str.split('/')
                exec_host[host] = cores
            info[key] = exec_host



#!/usr/bin/env python
'''class for parsing PBS files'''

import os
import re
from vsc.event_logger import EventLogger
from vsc.pbs.job import PbsJob
from vsc.pbs.option_parser import PbsOptionParser


class PbsScriptParser(EventLogger):
    '''Parser for PBS torque job files'''

    def __init__(self, config, event_defs, pbs_directive='#PBS'):
        '''Constructor'''
        super(PbsScriptParser, self).__init__(event_defs)
        self._config = config
        self._job = PbsJob(self._config)
        self._pbs_directive = pbs_directive
        regex = r'\s*{0}\s+(.+)$'.format(pbs_directive)
        self._pbs_re = re.compile(regex)
        regex = r'\s+{0}\s+(.+)$'.format(pbs_directive)
        self._pbs_indented_re = re.compile(regex)
        regex = r'\s*{0}\s+(.+)$'.format(pbs_directive)
        self._pbs_extract_re = re.compile(regex)
        regex = r'\s*-[A-Za-z]'
        self._pbs_option_re = re.compile(regex)
        self._pbs_option_parser = PbsOptionParser(self._config, event_defs,
                                                  self._job)
        self._state = None
        self._line_nr = 0
        self._pbs = []
        self._script_first_line_nr = None

    @property
    def script_first_line_nr(self):
        '''return the first line number of the Bash script part'''
        return self._script_first_line_nr

    def parse_file(self, pbs_file):
        '''parse a PBS file'''
        self._job.name = os.path.basename(pbs_file.name)
        self._state = 'start'
        self._line_nr = 0
        for line in pbs_file:
            self._line_nr += 1
            self.check_encoding(line)
            if self.is_comment(line):
                continue
            if len(line.strip()) == 0:
                continue
            if self._state == 'start':
                self.parse_shebang(line)
            elif self._state == 'pbs':
                self.parse_pbs(line)
            else:
                self.parse_script(line)
        if self._state == 'start':
            self.reg_event('no_script')
        elif self._state == 'pbs':
            self.reg_event('no_script')

    @property
    def job(self):
        '''returns a PbsJob object representing the job script'''
        return self._job

    def check_encoding(self, line):
        '''checks ASCII encoding and line endings'''
        try:
            line.encode('ascii')
        except UnicodeDecodeError:
            self.reg_event('non_ascii')
        if line.endswith('\r\n'):
            self.reg_event('dos_format')
        if line.endswith('\r'):
            self.reg_event('mac_format')

    def is_comment(self, line):
        '''returns True if the line is a comment'''
        return (re.match(r'\s*#', line) and
                not (self.is_shebang(line) or
                     self.is_spaced_pbs(line) or
                     self.is_indented_pbs(line) or
                     self.is_pbs(line)))

    def is_shebang(self, line):
        '''returns True if the line is a shebang'''
        return line.startswith('#!')

    def is_spaced_pbs(self, line):
        '''checks whether an extra space is added between hash and text
           in PBS directive, only for default directive string'''
        return self._pbs_directive == '#PBS' and re.match(r'\s*#\s+PBS\s+',
                                                          line)

    def is_pbs(self, line):
        '''returns True if the line is a PBS directive'''
        return self._pbs_re.match(line)

    def is_indented_pbs(self, line):
        '''returns True if the line contains an indented PBS directive'''
        return self._pbs_indented_re.match(line)

    def parse_shebang(self, line):
        '''parse shebang part of PBS file'''
        if self.is_shebang(line):
            self._job.shebang = line.strip()
            if self._line_nr > 1:
                self.reg_event('misplaced_shebang')
            self._state = 'pbs'
        else:
            self.reg_event('missing_shebang')
            self._state = 'pbs'
            self.parse_pbs(line)

    def parse_pbs(self, line):
        '''parse PBS directives part of a PBS file'''
        if self.is_shebang(line):
            self.reg_event('misplaced_shebang')
        elif self.is_spaced_pbs(line):
            self.reg_event('space_in_pbs_dir')
        elif self.is_pbs(line):
            if self.is_indented_pbs(line):
                self.reg_event('indented_pbs_dir')
            match = self._pbs_extract_re.match(line)
            if match:
                option = match.group(1)
                if self._pbs_option_re.match(option):
                    self._pbs_option_parser.parse_args(option)
                    self.merge_events(self._pbs_option_parser.events)
                else:
                    self.reg_event('malformed_pbs_dir')
            else:
                self.reg_event('malformed_pbs_dir')
        else:
            self._state = 'script'
            if not self._script_first_line_nr:
                self._script_first_line_nr = self._line_nr
            self.parse_script(line)

    def parse_script(self, line):
        '''parse shell script part of a PBS file'''
        if self.is_shebang(line):
            self.reg_event('misplaced_shebang')
        if self.is_pbs(line):
            self.reg_event('misplace_pbs_dir')
        self._job.add_script_line(self._line_nr, line)



'''Module to parse torque logs'''

from datetime import datetime, timedelta
import os
import sys

from vsc.pbs.job import PbsJob
from vsc.pbs.job_event import PbsJobEvent


class PbsLogParserError(Exception):
    '''Exception to denote a parse error'''

    def __init__(self, msg):
        self._msg = msg

    def __str__(self):
        return self._msg


class PbsLogParser(object):
    '''Implements torque log parser'''

    def __init__(self, config):
        '''Constructor'''
        self._config = config
        self._jobs = dict()

    def parse(self, start_date, end_date):
        '''Parse log information from the given start date to the end
        date, inclusive'''
        start_date = datetime.strptime(start_date, '%Y%m%d')
        end_date = datetime.strptime(end_date, '%Y%m%d')
        date = start_date
        delta_time = timedelta(days=1)
        while date <= end_date:
            file_name = os.path.join(self._config['log_dir'],
                                     date.strftime('%Y%m%d'))
            if os.path.exists(file_name):

                self.parse_file(file_name)
            else:
                sys.stderr.write("no log file '{0}'\n".format(file_name))
            date += delta_time

    def parse_file(self, file_name):
        '''Parse the specified log'''
        with open(file_name, 'r') as pbs_file:
            line_nr = 0
            for line in pbs_file:
                line_nr += 1
                line = line.rstrip()
                if not line:
                    continue
                try:
                    time_stamp, event_type, job_id, info_str = line.split(';')
                    if job_id not in self._jobs:
                        self._jobs[job_id] = PbsJob(self._config, job_id)
                    event = PbsJobEvent(time_stamp, event_type, info_str)
                    self._jobs[job_id].add_event(event)
                except:
                    msg = 'problem on line {0} in {1}'.format(line,
                                                              file_name)
                    raise PbsLogParserError(msg)

    @property
    def jobs(self):
        '''return the jobs that were parsed as a dictionary that maps
        job IDs to job object'''
        return self._jobs



'''module for dealing with Moab's checkjob output'''

import re


class CheckjobParser(object):
    '''Parser class for Moab checkjob ouptut'''

    def __init__(self):
        self._account_re = re.compile(r'account:(\S+)')

    def parse(self, job, checkjob_str):
        '''parse checkjob output, and return relevant status'''
        for line in checkjob_str.split('\n'):
            match = self._account_re.search(line)
            if match:
                job.account = match.group(1)
                continue
            if line.startswith('Holds:'):
                _, hold_str = line.strip().split(' ', 1)
                hold_strs = hold_str.strip().split(',')
                for hold in hold_strs:
                    job.add_hold(hold)



#!/usr/bin/env python
'''module for dealing with events'''

class UndefinedEventError(Exception):

    def __init__(self, msg):
        super(UndefinedEventError, self).__init__(msg)


class EventLogger(object):
    '''base class for classes that need to register events'''

    def __init__(self, event_defs, context='file'):
        '''Constructor, to be called by dervied classes'''
        super(EventLogger, self).__init__()
        self._event_defs = event_defs
        self._events = []
        self._context = context

    @property
    def context(self):
        '''return context the EventLogger is operating in'''
        return self._context

    @context.setter
    def context(self, context):
        '''set the context for the EventLogger'''
        self._context = context

    @property
    def events(self):
        '''return events generated during parsing'''
        return self._events

    def reg_event(self, event, extra={}, line_nr=None):
        '''register a event'''
        if event not in self._event_defs:
            msg = "event '{0}' is undefined".format(event)
            raise UndefinedEventError(msg)
        if self.context == 'file':
            line_nr = line_nr if line_nr else self._line_nr
            self._events.append({'id': event,
                                 'line': line_nr,
                                 'extra': extra})
        else:
            self._events.append({'id': event,
                                 'line': None,
                                 'extra': extra})

    def merge_events(self, events):
        '''merge events'''
        for event in events:
            self.reg_event(event['id'], event['extra'], event['line'])

    @property
    def nr_errors(self):
        '''return number of errors logged so far'''
        nr_errors = 0
        for event in self._events:
            if self._event_defs[event['id']]['category'] == 'error':
                nr_errors += 1
        return nr_errors

    @property
    def nr_warnings(self):
        '''return number of warnings logged so far'''
        nr_warnings = 0
        for event in self._events:
            if self._event_defs[event['id']]['category'] == 'warning':
                nr_warnings += 1
        return nr_warnings



'''module containing logic to compute the credit cost of a job'''

import sqlite3


class QuoteCalculator(object):
    '''class implementing a calculator of job costs'''

    def __init__(self, config):
        '''constructor that takes a configuration'''
        self._config = config

    def determine_node_types(self, node_spec, partition):
        '''determine the relevant type of a node to base to cost
           calculation on'''
        if 'properties' in node_spec:
            node_types = node_spec['properties']
        else:
            node_types = []
            with sqlite3.connect(self._config['cluster_db']) as conn:
                cursor = conn.cursor()
                result = cursor.execute(
                    '''SELECT DISTINCT p.property
                           FROM properties as p, nodes as n,
                                partitions as p
                           WHERE p.node_id = n.node_id AND
                                 n.partition_id = p.partition_id AND
                                 p.partition_name = ?''',
                    (partition, ))
                for row in result:
                    node_types.append(row[0])
        return node_types

    def compute(self, job):
        '''compute the cost in credits of the given job'''
        type_costs = self._config['node_type_credits']
        walltime = float(job.resource_specs['walltime'])
        node_rate = self._config['node_rate']
        partition = job.resource_specs['partition']
        resource_specs = job.resource_specs['nodes']
        cost = 0.0
        for resource_spec in resource_specs:
            nr_nodes = resource_spec['nodes']
            node_types = self.determine_node_types(resource_spec, partition)
            node_cost = -1.0
            for node_type in node_types:
                if node_type in type_costs:
                    if type_costs[node_type] > node_cost:
                        node_cost = type_costs[node_type]
            if node_cost < 0.0:
                node_cost = type_costs[job.resource_spec('partition')]
            cost += nr_nodes*node_cost
        return cost*walltime*node_rate



'''Module to parse PBS torque's pbsnodes command output'''

import re, sys

from vsc.pbs.node import NodeStatus

class PbsnodesParser:
    '''Implements a parser for pbsnodes output'''

    def __init__(self, is_verbose=False):
        '''Constructor'''
        self._is_verbose = is_verbose

    def parse(self, node_output):
        '''parse output as produced by pbsnodes, returns a list of nodes'''
        nodes = []
        state = 'init'
        for line in node_output.split('\n'):
            if state == 'init' and re.match(r'^\w', line):
                node_str = line
                state = 'in_node'
            elif state == 'in_node':
                if len(line.strip()):
                    node_str += '\n' + line
                else:
                    try:
                        nodes.append(self.parse_node(node_str.strip()))
                    except Exception as e:
                        msg = "### warning: error parsing: '{}'\n"
                        sys.stderr.write(msg.format(str(e)))
                    state = 'init'
        return nodes

    def parse_file(self, node_file):
        '''parse a file that contains pbsnodes output'''
        node_output = ''.join(node_file.readlines())
        return self.parse(node_output)

    def parse_jobs(self, jobs_str):
        job_info = {}
        job_strs = jobs_str.split(' ')
        for job_str in job_strs:
            if '(' in job_str:
                job_id, info_str = job_str.split('(')
                job_info[job_id] = {}
                info_strs = info_str[:-2]
                for info_str in info_strs.split(','):
                    try:
                        key, value = info_str.split('=')
                        job_info[job_id][key] = value
                    except ValueError:
                        if self._is_verbose:
                            msg = '### warnig: no value for {0}\n'
                            sys.stderr.write(msg.format(key))
                        job_info[job_id][info_str] = None
        return job_info
        
    def parse_status(self, node_status, status_str):
        pos = status_str.find(',')
        while pos >= 0:
            if status_str.startswith('jobs='):
                last_pos = status_str.rfind(')')
                if last_pos == -1:
                    jobs = {}
                    status_str = status_str[pos + 1:]
                else:
                    start_pos = status_str.find('=') + 1
                    jobs_str = status_str[start_pos:last_pos + 1]
                    jobs = self.parse_jobs(jobs_str)
                    status_str = status_str[last_pos + 2:]
                node_status.status['job_info'] = jobs
                pos = status_str.find(',')

            else:
                try:
                    key, value = status_str[:pos].split('=')
                    node_status.status[key] = value
                except ValueError:
                    if self._is_verbose:
                        msg = '### warning: no value for {0} on {1}\n'
                        sys.stderr.write(msg.format(key,
                                                    node_status.hostname))
                    node_status.status[status_str[:pos]] = None
                status_str = status_str[pos + 1:]
                pos = status_str.find(',')

    def parse_gpu_status(self, gpu_status_str):
        '''parse the gpu_status entry'''
        gpu_status = list()
        gpu_strs = gpu_status_str.split(',')
        for gpu_str in gpu_strs:
            match = re.match(r'^gpu\[(\d+)\]=(.+)$', gpu_str)
            if match is None:
                msg = "### warning: can not parse gpu_status '{}'\n"
                # sys.stderr.write(msg.format(gpu_str))
                continue
            gpu_info = {'gpu_nr': int(match.group(1))}
            for info in match.group(2).split(';'):
                key, value = info.split('=')
                gpu_info[key] = value
            gpu_status.append(gpu_info)
        return gpu_status

    def parse_node(self, node_str):
        '''parse a string containing pbsnodes information of single node'''
        lines = node_str.split('\n')
        hostname = lines.pop(0).strip()
        node_status = NodeStatus(hostname)
        for line in (l.strip() for l in lines):
            if line.startswith('np = '):
                _, np_str = line.split(' = ')
                node_status.np = int(np_str)
            elif line.startswith('properties = '):
                _, properties_str = line.split(' = ')
                node_status.properties = properties_str.split(',')
            elif line.startswith('status = '):
                _, status_str = line.split(' = ')
                node_status.status = {}
                if 'message=' in status_str:
                    match = re.search(r'message=(.*?)(?:,\w+=|$)',
                                      status_str)
                    message = match.group(1)
                    node_status.status['message'] = message
                    repl = 'message={0},'.format(message)
                    status_str = status_str.replace(repl, '')
                    if self._is_verbose:
                        msg = '### warning: message {0} on node {1}\n'
                        sys.stderr.write(msg.format(message, hostname))
                self.parse_status(node_status, status_str)
            elif line.startswith('ntype = '):
                _, node_status.ntype = line.split(' = ')
            elif line.startswith('state = '):
                _, node_status.state = line.split(' = ')
            elif line.startswith('note = '):
                _, node_status.note = line.split(' = ')
            elif line.startswith('jobs = '):
                _, job_str = line.split(' = ')
                node_status.jobs = {}
                for job_item in job_str.split(','):
                    core, job = job_item.split('/')
                    node_status.jobs[core] = job
            elif line.startswith('gpu_status = '):
                _, gpu_status_str = line.split(' = ')
                for gpu_status in self.parse_gpu_status(gpu_status_str):
                    node_status.add_gpu_status(gpu_status)
        return node_status



#!/usr/bin/env python

from configparser import ConfigParser
import datetime
import plotly.plotly as py
from plotly.graph_objs import Font

def create_annotations():
    timestamp = datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')
    anno_text = 'time: {0}'.format(timestamp)
    annotations = [
        dict(
            text=anno_text,  # set annotation text
            showarrow=False, # remove arrow 
            xref='paper',  # use paper coords
            yref='paper',  #  for both coordinates
            x=0.95,  # position's x-coord
            y=1.10,  #   and y-coord
            font=Font(size=16),    # increase font size (default is 12)
            bgcolor='#FFFFFF',     # white background
            borderpad=4            # space bt. border and text (in px)
        )
    ]
    return annotations

def sign_in(file_name):
    '''read credentials from config file, and log in to Plot.ly'''
    conf_parser = ConfigParser()
    conf_parser.read(file_name)
    username = conf_parser.get('authentication', 'username')
    api_key = conf_parser.get('authentication', 'api_key')
    py.sign_in(username, api_key)



'''classes and utilities to represent and manipulate Adaptive MAM accounts'''

class MamAccount(object):
    '''class representing a MAM account'''

    def __init__(self, acc_id, name):
        '''constructor taking account ID, name'''
        self._id = acc_id
        self._name = name
        self._available = None
        self._allocated = None

    @property
    def account_id(self):
        '''returns MAM account ID'''
        return self._id

    @property
    def name(self):
        '''returns MAM account name'''
        return self._name

    @property
    def available_credits(self):
        '''returns number of credits available on this account'''
        return self._available

    @available_credits.setter
    def available_credits(self, credits):
        '''set the number of credits available on this account'''
        self._available = float(credits)

    @property
    def allocated_credits(self):
        '''returns total allocated credits for this account'''
        return self._allocated

    @allocated_credits.setter
    def allocated_credits(self, credits):
        '''set the total number of credits allocated on this account'''
        self._allocated = float(credits)

    def __str__(self):
        account_str = '{0} ({1}):'.format(self.name, self.account_id)
        account_str += '\n  available: {0:.2f}'.format(self.available_credits)
        account_str += '\n  allocated: {0:.2f}'.format(self.allocated_credits)
        return account_str



#!/usr/bin/env python
'''Module containing a number of useful functions'''

class InvalidWalltimeError(Exception):
    '''thrown when an invalid walltime format is used'''

    def __init__(self, msg):
        super(InvalidWalltimeError, self).__init__()
        self.message = msg

    def __str__(self):
        return self.message


class InvalidSizeError(Exception):
    '''thrown when an invalid size format is used'''

    def __init__(self, msg):
        super(InvalidSizeError, self).__init__()
        self.message = msg

    def __str__(self):
        return self.message


import math, re

def walltime2seconds(time_str):
    '''converts walltime [[[DD:]HHH?:]MM:]SS+ to seconds

    >>> walltime2seconds('1234')
    1234
    >>> walltime2seconds('13:12')
    792
    >>> walltime2seconds('3:02:45')
    10965
    >>> walltime2seconds('3:01:02:03')
    262923
    >>> walltime2seconds('1:2:3')
    Traceback (most recent call last):
        ...
    InvalidWalltimeError: '1:2:3' is invalid
    >>> walltime2seconds('1-02-03')
    Traceback (most recent call last):
        ...
    InvalidWalltimeError: '1-02-03' is invalid
    '''
    time_str = time_str.strip()
    match = re.match(r'^(\d+)$', time_str)
    if match:
        return int(match.group(1))
    match = re.match(r'^(\d+):(\d{2})$', time_str)
    if match and int(match.group(2)) < 60:
        return 60*int(match.group(1)) + int(match.group(2))
    match = re.match(r'^(\d+):(\d{2}):(\d{2})$', time_str)
    if match and int(match.group(2)) < 60 and int(match.group(3)) < 60:
        return (3600*int(match.group(1)) + 60*int(match.group(2)) +
                int(match.group(3)))
    match = re.match(r'^(\d+):(\d{2}):(\d{2}):(\d{2})$', time_str)
    if match and (int(match.group(2)) < 24 and
                  int(match.group(3)) < 60 and
                  int(match.group(4)) < 60):
        return (24*3600*int(match.group(1)) + 3600*int(match.group(2)) +
                60*int(match.group(3)) + int(match.group(4)))
    raise InvalidWalltimeError("'{0}' is invalid".format(time_str))

def seconds2walltime(seconds):
    '''convert a time in seconds to [[HHH:[MM:]SS

    >>> seconds2walltime(12)
    '00:00:12'
    >>> seconds2walltime(123)
    '00:02:03'
    >>> seconds2walltime(1234)
    '00:20:34'
    >>> seconds2walltime(12345)
    '03:25:45'
    '''
    seconds = int(seconds)
    secs = seconds % 60
    seconds = seconds//60
    mins = seconds % 60
    hours = seconds//60
    return '{h:02d}:{m:02d}:{s:02d}'.format(h=hours, m=mins, s=secs)

def size2bytes(amount, order=None):
    '''given a number and and order, compute size
    >>> size2bytes(12, 'k')
    12288
    >>> size2bytes(12, 't')
    13194139533312
    >>> size2bytes(12, None)
    12
    >>> size2bytes(12, 'q')
    Traceback (most recent call last):
        ...
    InvalidSizeError: 'q' is not a valid order of magnitude
    >>> size2bytes('size', 't')
    Traceback (most recent call last):
        ...
    InvalidSizeError: 'size' is not an integer
    >>> size2bytes('12kb')
    12288
    >>> size2bytes('12 tw')
    13194139533312
    '''
    conversion = {
        'k': 1024,
        'm': 1024**2,
        'g': 1024**3,
        't': 1024**4,
        None: 1,
    }
    if (type(amount) == str and ('b' in amount or 'w' in amount) and
        order is None):
        match = re.match(r'(\d+)\s*([kmgt]?)(?:b|w)$', amount)
        if match:
            amount = match.group(1)
            order = match.group(2)
        else:
            raise InvalidSizeError("'{0}' is not an integer".format(amount))
    try:
        if order:
            return int(amount)*conversion[order]
        else:
            return int(amount)
    except ValueError:
        raise InvalidSizeError("'{0}' is not an integer".format(amount))
    except KeyError:
        raise InvalidSizeError("'{0}' is not a valid order "
                               "of magnitude".format(order))

def bytes2size(bytes, unit, no_unit=False, fraction=False):
    '''Conbert a number of bytes to the given unit (kb, mb, gb, tb)
       >>> bytes2size(34320, 'kb')
       '34kb'
       >>> bytes2size(12884463294, 'GB')
       '12GB'
       >>> bytes2size(12884463294, 'GB', no_unit=True)
       '12'
       >>> bytes2size(1297, 'kb', fraction=True)
       '1.3kb'
    '''
    conversion = {
        'b': 1.0,
        'kb': 1024.0,
        'mb': 1024.0**2,
        'gb': 1024.0**3,
        'tb': 1024.0**4,
    }
    if unit.lower() in conversion:
        mem = bytes/conversion[unit.lower()]
        if  fraction:
            return '{0:.1f}{1}'.format(mem, '' if no_unit else unit)
        else:
            mem = int(math.ceil(mem))
            return '{0:d}{1}'.format(mem, '' if no_unit else unit)
    else:
        raise InvalidSizeError("'{0}' is not a valid unit".format(unit))
    

def hostname2rackinfo(hostname):
    '''Determine rack number, IRU and node number from hostname'''
    match = re.match(r'r(\d+)i(\d+)n(\d+)', hostname)
    if match:
        return int(match.group(1)), int(match.group(2)), int(match.group(3))
    else:
        return None, None, None


def core_specs2count(core_spec_str):
    '''Determine the number of cores from a speciciatoin such as, e.g.,
    0-5,7,9-11

    >>> core_specs2count('0-5')
    6
    >>> core_specs2count('1-9,15-18')
    13
    >>> core_specs2count('5')
    1
    >>> core_specs2count('0-4,7,9-14,3')
    13
    '''
    core_count = 0
    for core_spec in core_spec_str.split(','):
        if core_spec.isdigit():
            core_count += 1
        elif '-' in core_spec:
            low, high = core_spec.split('-')
            core_count += int(high) - int(low) + 1
    return core_count

if __name__ == '__main__':
    import doctest
    doctest.testmod()




#!/usr/bin/env python

class InconsistantAttributesError(Exception):
    '''Exception indicating inconsistant attributes in the job status
       constructor'''

    def __init__(self, msg):
        '''Constructor'''
        super(InconsistantAttributesError, self).__init__()
        self.message = msg


class JobStatus(object):
    '''Moab job status'''

    def __init__(self, id, username, state, procs, time, date):
        '''Job status constructor'''
        self._id = id
        self._username = username
        self._state = state
        self._procs = procs
        self._time = time
        self._date = date
        self._account = None
        self._holds = []

    @property
    def id(self):
        '''Returns job ID'''
        return self._id

    @property
    def username(self):
        '''Returns user name'''
        return self._username

    @property
    def state(self):
        '''Returns job state'''
        return self._state

    @property
    def procs(self):
        '''Returns number of processes'''
        return self._procs

    @property
    def remaining(self):
        '''Returns time remaining for a running job'''
        if self._state == 'Running':
            return self._time
        else:
            raise InconsistantAttributesError('non-running job has no remaining time')
        
    @property
    def starttime(self):
        '''Returns start time of a running job'''
        if self._state == 'Running':
            return self._date
        else:
            raise InconsistantAttributesError('non-running job has no starting time')
        
    @property
    def wclimit(self):
        '''Returns walltime limit of a non-running job'''
        if self._state != 'Running':
            return self._time
        else:
            raise InconsistantAttributesError('running job has no wall clock limit')
        
    @property
    def queuetime(self):
        '''Returns time non-running job was queued'''
        if self._state != 'Running':
            return self._date
        else:
            raise InconsistantAttributesError('running job has no queue time')

    @property
    def account(self):
        '''Return the account specified for charging, if any'''
        return self._account

    @account.setter
    def account(self, account):
        '''Set the account specified for charging'''
        self._account = account

    @property
    def holds(self):
        '''Returns list of holds in place on job, may be empty if no
           holds are in place'''
        return self._holds

    def add_hold(self, hold):
        '''Add a hold to the job status'''
        self._holds.append(hold)

    def __str__(self):
        '''Returns string representation of a job status'''
        job_str = 'ID {0}:'.format(self.id)
        job_str += '\n\tuser: {0}'.format(self.username)
        job_str += '\n\tstate: {0}'.format(self.state)
        job_str += '\n\tprocs: {0}'.format(self.procs)
        if self.state == 'Running':
            job_str += '\n\tremaining: {0}'.format(self.remaining)
        else:
            job_str += '\n\twclimit: {0}'.format(self.wclimit)
        if self.state == 'Running':
            job_str += '\n\tstarted: {0}'.format(self.starttime)
        else:
            job_str += '\n\tqueued: {0}'.format(self.queuetime)
        return job_str



'''Module implementing utilities to analyze PBS torque job log files'''

from collections import namedtuple
from datetime import date, datetime
import json
from operator import itemgetter
import pandas as pd

from vsc.pbs.log import PbsLogParser
from vsc.utils import seconds2walltime, bytes2size


class AnalysisError(Exception):
    '''Signals a problem in the analysis'''

    def __init__(self, msg):
        self._msg = msg

    def __str__(self):
        return self._msg


class PbsLogAnalysis(object):
    '''class that wraps log analysis functionality'''

    default_job_columns = [
        'start', 'end', 'job_id', 'user', 'state', 'partition',
        'used_mem', 'used_walltime', 'spec_walltime', 'nodes', 'ppn',
        'hosts', 'exit_status',
    ]
    JobTuple = namedtuple('JobTuple', default_job_columns)
    default_host_columns = ['job_id', 'host', 'cores']
    HostTuple = namedtuple('HostTuple', default_host_columns)
    time_fmt = '%Y-%m-%d %H:%M:%S'

    def __init__(self, cfg_file_name):
        '''Constructor that takes the name of a JSON configuration file'''
        with open(cfg_file_name, 'r') as json_file:
            self._config = json.load(json_file)
        self._jobs = None
        self._df_jobs = None
        self._df_hosts = None

    def prepare(self, start=None, end=None):
        '''Prepare analysis by parsing log files and acquiring
        information'''
        pbs_parser = PbsLogParser(self._config)
        if not end:
            end_date = date.today()
            end = end_date.strftime('%Y%m%d')
        if not start:
            start = self._config.log_start
        pbs_parser.parse(start, end)
        self._jobs = pbs_parser.jobs
        self._df_jobs, self._df_hosts = self._jobs_to_dataframes()

    @property
    def jobs(self):
        '''return jobs in the analysis'''
        return self._jobs

    @property
    def jobs_df(self):
        '''return jobs pandas data frame'''
        return self._df_jobs

    @property
    def hosts_df(self):
        '''return hosts pandas data frame'''
        return self._df_hosts

    @staticmethod
    def _job_to_tuple(job):
        '''returns a tuple with the job information that will be inserted
        into a pandas data frame'''
        if job.start:
            start_time = job.start.strftime(PbsLogAnalysis.time_fmt)
        else:
            start_time = '1970-01-01 00:00:00'
        if job.end:
            end_time = job.end.strftime(PbsLogAnalysis.time_fmt)
        else:
            end_time = datetime.now().strftime(PbsLogAnalysis.time_fmt)
        nodes = job.resource_spec('nodes')[0]['nodes']
        if 'ppn' in job.resource_spec('nodes')[0]:
            ppn = job.resource_spec('nodes')[0]['ppn']
        else:
            ppn = None
        if job.resource_used('mem'):
            mem_used = float(bytes2size(job.resource_used('mem'), 'gb',
                                        no_unit=True, fraction=True))
        else:
            mem_used = None
        if job.resource_used('walltime'):
            walltime = seconds2walltime(job.resource_used('walltime'))
        else:
            walltime = None
        spec_walltime = seconds2walltime(job.resource_spec('walltime'))
        if job.exec_host:
            hosts = ' '.join(list(job.exec_host.keys()))
        else:
            hosts = None
        return PbsLogAnalysis.JobTuple(
                start=start_time,
                end=end_time,
                job_id=job.job_id,
                user=job.user,
                state=job.state,
                partition=job.partition,
                used_mem=mem_used,
                used_walltime=walltime,
                spec_walltime=spec_walltime,
                nodes=nodes,
                ppn=ppn,
                hosts=hosts,
                exit_status=job.exit_status
                )

    @staticmethod
    def _exec_host_to_tuples(job):
        '''returns a tuple with the exec_host information that will be
        inserted into a pandas data frame'''
        tuples = []
        if job.exec_host:
            for host in sorted(job.exec_host.keys()):
                tuples.append((job.job_id, host, job.exec_host[host]))
        return tuples

    def _jobs_to_dataframes(self):
        '''create pandas DataFrames out of a dictionary of jobs, the first
        contains the job information, the second keeps track of job-node
        associations'''
        job_data = []
        host_data = []
        for job_id, job in self._jobs.items():
            if job.has_end_event() or job.has_start_event():
                job_data.append(PbsLogAnalysis._job_to_tuple(job))
                host_data.extend(PbsLogAnalysis._exec_host_to_tuples(job))
        df_jobs = pd.DataFrame(sorted(job_data, key=itemgetter(2)),
                               columns=PbsLogAnalysis.default_job_columns)

        def time_conv(time):
            return pd.datetime.strptime(time, PbsLogAnalysis.time_fmt)
        df_jobs['start'] = df_jobs['start'].map(time_conv)
        df_jobs['end'] = df_jobs['end'].map(time_conv)
        df_jobs.ppn = df_jobs.ppn.fillna(-1.0).astype(int)
        df_jobs.exit_status = df_jobs.exit_status.fillna(-1024.0).astype(int)
        df_hosts = pd.DataFrame(host_data,
                                columns=PbsLogAnalysis.default_host_columns)
        return df_jobs, df_hosts

    def running_jobs(self, start_time=None, end_time=None, at_time=None):
        '''returns a pandas DataFrame containing the jobs that were
        running within the time interval specified by start and end
        time'''
        jobs = self.jobs_df
        if at_time:
            return jobs[(jobs.start < at_time) & (at_time < jobs.end)]
        elif start_time and end_time:
            return jobs[(jobs.start < end_time) & (jobs.end > start_time)]
        else:
            raise AnalysisError('either specify interval, or time at')



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name='vsc-tools-lib-gjbex',
    version='0.9.1',
    author="Geert Jan Bex",
    author_email="geertjan.bex@uhasselt.be",
    description="Python library for interacting with PBS, Adaptive Moab and MAM",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/gjbex/vsc-tools-lib",
    packages=setuptools.find_packages(),
    test_suite='nose.collector',
    tests_require=['nose'],
    install_requires=[
        'fuzzywuzzy',
        'validate_email',
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
)



'''Miscellaneous functions for dealing the PBS data'''

def compute_features(node):
    '''Compute features for a node'''
    features = []
    if 'thinking' in node.properties:
        if node.memory <= 64*1024**3:
            features.append('mem64')
        else:
            features.append('mem128')
    return features

def compute_partition(node, partitions):
    '''Compute partition for a node based on its properties'''
    partition_id = None
    if type(partitions) == dict:
        for partition, id in partitions.items():
            if node.has_property(partition):
                partition_id = id
                break
    elif type(partitions) == list:
        for partition in partitions:
            if node.has_property(partition):
                partition_id = partition
                break
    return partition_id



'''module to represent and manipulate PBS jobs'''

from datetime import datetime
import operator
import os


class PbsJob(object):
    '''Class representing a PBS job'''

    def __init__(self, config, job_id=None):
        '''Constructor for a PBS job object'''
        self._id = job_id
        self._name = None
        self._user = None
        self._state = None
        self._resources_used = {}
        self._exec_hosts = None
        self._partition = None
        self._resource_specs = {
            'pmem': config['default_pmem'],
            'partition': config['default_partition'],
            'qos': config['default_qos'],
            'nodes': [{'nodes': config['default_nodes'],
                       'ppn': config['default_ppn'],
                       'properties': [], }],
            'walltime': config['default_walltime'],
            'features': [],
        }
        self._has_default_pmem = True
        self._queue = config['default_queue']
        self._exit_status = None
        self._project = None
        _, host, _, _, _ = os.uname()
        cwd = os.getcwd()
        self._io_specs = {
            'keep': config['default_keep'],
            'join': config['default_join'],
            'error': {'host': host, 'path': cwd},
            'output': {'host': host, 'path': cwd},
        }
        try:
            default_mail_addr = os.getlogin()
        except OSError:
            default_mail_addr = ''
        self._mail_specs = {
            'events': config['default_mail_events'],
            'addresses': [default_mail_addr]
        }
        self._shebang = None
        self._script = []
        self._is_time_limit_set = False
        self._events = []

    @property
    def job_id(self):
        '''Returns job ID, None for jobs that are not queued'''
        return self._id

    @property
    def name(self):
        '''returns the job's name, None if not set'''
        return self._name

    @name.setter
    def name(self, name):
        '''Set the job's name'''
        self._name = name

    @property
    def user(self):
        '''Returs the user ID of the job's woner, if any'''
        return self._user

    @user.setter
    def user(self, user):
        '''Sets the job's owner's user ID'''
        self._user = user

    @property
    def state(self):
        '''Returs the job's state'''
        return self._state

    @state.setter
    def state(self, state):
        '''Sets the job's state'''
        self._state = state

    @property
    def partition(self):
        '''Returs the job's partition'''
        return self._partition

    @partition.setter
    def partition(self, partition):
        '''Sets the job's partition'''
        self._partition = partition

    @property
    def exec_host(self):
        '''Returns a map with exec hosts for a running job'''
        return self._exec_hosts

    @exec_host.setter
    def exec_host(self, exec_host):
        '''Sets the hosts a job is executing on'''
        self._exec_hosts = exec_host

    @property
    def project(self):
        '''Returns the job's project name, None if not set'''
        return self._project

    @project.setter
    def project(self, name):
        '''Set the job's project name'''
        self._project = name

    @property
    def queue(self):
        '''returns the job's queue name, None if not set'''
        return self._queue

    @queue.setter
    def queue(self, name):
        '''Set the job's queue name'''
        self._queue = name

    @property
    def exit_status(self):
        '''returns the job's exit status, or None'''
        return self._exit_status

    @exit_status.setter
    def exit_status(self, value):
        '''Set the job's exit status'''
        self._exit_status = value

    @property
    def resource_specs(self):
        '''returns the job's resource specifications'''
        return self._resource_specs

    def resource_spec(self, spec):
        '''returns the resource specification specified, None if that
           was not specified'''
        if spec in self._resource_specs:
            return self._resource_specs[spec]
        else:
            return None

    def add_resource_spec(self, key, value):
        self._resource_specs[key] = value

    def add_resource_specs(self, resource_specs):
        '''Add resources to specification'''
        for key, value in list(resource_specs.items()):
            self.add_resource_spec(key, value)

    @property
    def resources_used(self):
        '''returns the job's resource usage'''
        return self._resources_used

    def resource_used(self, key):
        if key in self._resources_used:
            return self._resources_used[key]
        else:
            return None

    def add_resources_used(self, resources_used):
        '''Add resources to used list'''
        for key, value in list(resources_used.items()):
            self.add_resource_used(key, value)

    def add_resource_used(self, key, value):
        self._resources_used[key] = value

    @property
    def has_default_pmem(self):
        '''returns True if the default value for pmem was not changed'''
        return self._has_default_pmem

    @property
    def mail_events(self):
        '''return mail events'''
        return self._mail_specs['events']

    @mail_events.setter
    def mail_events(self, events):
        '''set mail events for job'''
        self._mail_specs['events'] = events

    @property
    def mail_addresses(self):
        '''return mail addresses'''
        return self._mail_specs['addresses']

    @mail_addresses.setter
    def mail_addresses(self, addresses):
        '''set mail address(es) for job mail events'''
        if type(addresses) == list:
            self._mail_specs['addresses'] = addresses
        else:
            self._mail_specs['addresses'] = [addresses]

    @property
    def join(self):
        '''return I/O join'''
        return self._io_specs['join']

    @join.setter
    def join(self, join):
        '''set I/O join for job'''
        self._io_specs['join'] = join

    @property
    def keep(self):
        '''return I/O keep'''
        return self._io_specs['keep']

    @keep.setter
    def keep(self, keep):
        '''set I/O keep for job'''
        self._io_specs['keep'] = keep

    @property
    def error(self):
        '''return path for job's standard error'''
        return (self._io_specs['error']['host'],
                self._io_specs['error']['path'])

    def set_error(self, path, host=None):
        '''set path for job's standard error'''
        if not os.path.isabs(path):
            path = os.path.join(self._io_specs['error']['path'], path)
        self._io_specs['error']['path'] = path
        if host:
            self._io_specs['error']['host'] = host

    @property
    def output(self):
        '''return path for job's standard output'''
        return (self._io_specs['output']['host'],
                self._io_specs['output']['path'])

    def set_output(self, path, host=None):
        '''set path for job's standard output'''
        if not os.path.isabs(path):
            path = os.path.join(self._io_specs['output']['path'], path)
        self._io_specs['output']['path'] = path
        if host:
            self._io_specs['output']['host'] = host

    @property
    def shebang(self):
        '''returns the job's shebang, None if not set'''
        return self._shebang

    @shebang.setter
    def shebang(self, shebang):
        '''Set the job's shebang'''
        self._shebang = shebang

    @property
    def script(self):
        '''returns the actual job script, i.e., the command to be executed'''
        return '\n'.join([line[1] for line in self._script])

    def add_script_line(self, line_nr, line):
        '''adds a line to the script, retains line number information'''
        self._script.append((line_nr, line))

    @property
    def events(self):
        '''return event list for this job'''
        return sorted(self._events, key=operator.attrgetter('time_stamp'))

    def has_start_event(self):
        '''returns True if this job has a start event'''
        return any(event.is_start() for event in self._events)

    def has_end_event(self):
        '''returns True if this job has a end event'''
        return any(event.is_end() for event in self._events)

    def add_event(self, event):
        '''add event to the job'''
        self._events.append(event)
        event.update_job_info(self)

    @property
    def end_event(self):
        if self.has_end_event():
            return self.events[-1]
        else:
            return None

    @property
    def start(self):
        '''return the start datetime of the job, None if not started'''
        if self.has_start_event():
            for event in self.events[::-1]:
                if event.type == 'S':
                    return event.time_stamp
        else:
            return None

    @property
    def end(self):
        '''return the end datetime of the job, None if not started'''
        if self.has_end_event():
            return self.events[-1].time_stamp
        else:
            return None

    @property
    def queue_time(self):
        '''return queue time as reported by qstat'''
        try:
            return self._queue_time
        except:
            return None

    @queue_time.setter
    def queue_time(self, value):
        '''set queue time as reported by qstat'''
        if type(value) == str:
            self._queue_time = datetime.strptime(value, '%c')
        else:
            self._queue_time = value

    @property
    def start_time(self):
        '''return start time as reported by qstat'''
        try:
            return self._start_time
        except:
            return None
    @start_time.setter
    def start_time(self, value):
        '''set start time as reported by qstat'''
        if type(value) == str:
            self._start_time = datetime.strptime(value, '%c')
        else:
            self._start_time = value

    @property
    def time_in_queue(self):
        if self.state == 'Q':
            return (datetime.now() - self.queue_time).total_seconds()
        else:
            return (self.start_time - self.queue_time).total_seconds()

    @property
    def walltime_used(self):
        '''return the walltime already used by the job, 0 if queued'''
        if self.state == 'Q':
            return 0
        else:
            return self.resource_used('walltime')

    @property
    def walltime_remaining(self):
        ''' return walltime remaining for a job, requested walltime if queued'''
        if self.state == 'Q':
            return self.resource_spec('walltime')
        else:
            delta = self.resource_spec('walltime') - self.resource_used('walltime')
            return delta if delta >= 0 else 0

    def attrs_to_str(self):
        '''return job attributes as a string, mainly for debug purposes'''
        attr_str = ''
        attr_str += "name = '{0}'".format(self.name)
        attr_str += "\nproject = '{0}'".format(self.project)
        attr_str += "\nresources:"
        for resource_name, resource_spec in list(self.resource_specs.items()):
            if resource_name == 'nodes':
                for node_spec in resource_spec:
                    for f_name, f_val in list(node_spec.items()):
                        attr_str += "\n\t{0} = '{1}'".format(f_name, f_val)
            else:
                attr_str += "\n\t{0} = '{1}'".format(resource_name,
                                                     resource_spec)
        attr_str += "\nqueue = {0}".format(self.queue)
        attr_str += '\nI/O:'
        attr_str += "\n\tjoin = '{0}'".format(self.join)
        attr_str += "\n\tkeep = '{0}'".format(self.keep)
        attr_str += "\n\terror = '{0}'".format(self.error)
        attr_str += "\n\toutput = '{0}'".format(self.output)
        attr_str += "\nmail:"
        attr_str += "\n\tevents = '{0}'".format(self.mail_events)
        address_str = ','.join(self.mail_addresses)
        attr_str += "\n\taddresses = '{0}'".format(address_str)
        return attr_str



'''Module to represent cluster nodes'''

import re

import vsc.utils

class NodeStatus(object):
    '''Class representing node status as reported by pbsnodes'''

    def __init__(self, hostname):
        '''Constructor'''
        self._hostname = hostname
        self._state = None
        self._np = None
        self._properties = None
        self._ntype = None
        self._jobs = None
        self._status = None
        self._note = None
        self._memory = 0
        self._gpu_status = list()

    @property
    def hostname(self):
        '''returns node's hostname'''
        return self._hostname

    @property
    def memory(self):
        '''returns the node's memory in bytes as integer'''
        if self._memory == 0 and self._status:
            self._memory = vsc.utils.size2bytes(self._status['physmem'])
        return self._memory

    @property
    def cpuload(self):
        '''Returns load average, None if no status information was
           available'''
        if self._status and self._status['loadave']:
            loadave = float(self._status['loadave'])
            ncpus = int(self._status['ncpus'])
            return loadave/ncpus
        return None

    @property
    def memload(self):
        '''Returns the memory load, None if no status information was
           available'''
        if (self._status and self._status['availmem'] and
                self._status['physmem']):
            physmem = vsc.utils.size2bytes(self._status['physmem'])
            availmem = vsc.utils.size2bytes(self._status['availmem'])
            return 1.0 - float(availmem)/float(physmem)
        return None

    @property
    def state(self):
        '''returns node's state'''
        return self._state

    @property
    def np(self):
        '''returns node's np'''
        return self._np

    @property
    def properties(self):
        '''returns node's properties'''
        return self._properties

    def has_property(self, prop):
        '''returns True if the NodeStatus has the given property'''
        if self._properties:
            return prop in self._properties
        else:
            return False

    @property
    def ntype(self):
        '''returns node's ntype'''
        return self._ntype

    @property
    def jobs(self):
        '''returns node's jobs'''
        return self._jobs

    @property
    def job_ids(self):
        '''returns a set of jobs IDs, empty if node is free'''
        job_ids = set()
        if self.jobs:
            for job_id in self.jobs.values():
                match = re.match(r'(\d+)', job_id)
                if match:
                    job_ids.add(match.group(1))
        return job_ids

    @property
    def status(self):
        '''returns node's status'''
        return self._status

    @property
    def note(self):
        '''returns node's note'''
        return self._note

    @state.setter
    def state(self, state):
        '''set node's state'''
        self._state = state

    @np.setter
    def np(self, np):
        '''set node's np'''
        self._np = int(np)

    @properties.setter
    def properties(self, properties):
        '''set node's properties'''
        self._properties = properties

    @ntype.setter
    def ntype(self, ntype):
        '''set node's ntype'''
        self._ntype = ntype

    @jobs.setter
    def jobs(self, jobs):
        '''set node's jobs'''
        self._jobs = jobs

    @status.setter
    def status(self, status):
        '''set node's status'''
        self._status = status

    @note.setter
    def note(self, note):
        '''set node's note'''
        self._note = note

    @property
    def gpu_status(self):
        '''return GPU info, list of dicts, one per GPU, empty list if no GPUs
        in node'''
        return self._gpu_status

    def add_gpu_status(self, gpu_status):
        '''add a gpu_status dictionary to the Node object'''
        self._gpu_status.append(gpu_status)

    def __str__(self):
        '''returns string representation for node status'''
        node_str = self.hostname
        node_str += '\n\t{0} = {1}'.format('state', self.state)
        node_str += '\n\t{0} = {1}'.format('np', self.np)
        properties_str = ','.join(self.properties)
        node_str += '\n\t{0} = {1}'.format('properties', properties_str)
        node_str += '\n\t{0} = {1}'.format('ntype', self.ntype)
        if self.jobs:
            jobs_str = ','.join(
                ['{0}={1}'.format(k, v) for k, v in self.jobs.items()]
            )
            node_str += '\n\t{0} = {1}'.format('jobs', jobs_str)
        if self.status:
            status_str = ','.join(
                ['{0}={1}'.format(k, v) for k, v in self.status.items()]
            )
            node_str += '\n\t{0} = {1}'.format('status', status_str)
        if self.note:
            node_str += '\n\t{0} = {1}'.format('note', self.note)
        return node_str




'''utilities for interacting with Adaptive MAM's gbalance'''

import re, sys

from vsc.mam.account import MamAccount

class GbalanceParser(object):
    '''parser for gbalance output'''

    def __init__(self):
        '''constructor'''
        self._acc_re = re.compile(r'Account=([^,]+),?')
        self._old_re = re.compile(r'(\w+)')

    def parse_file(self, gbalance_file):
        '''parse a file that gbalance pbsnodes output'''
        gbalance_output = ''.join(gbalance_file.readlines())
        return self.parse(gbalance_output)

    def parse(self, gbalance_output):
        '''parse the output of gbalance and return account information'''
        accounts = {}
        for line in gbalance_output.split('\n'):
            if len(line.strip()) == 0:
                continue
            if line.startswith('Id'):
                field_names = [x.strip() for x in line.split()]
                continue
            if line.startswith('--'):
                field_widths = [len(x) for x in line.split()]
                continue
            account_info = {}
            offset = 0
            for field_width, field_name in zip(field_widths, field_names):
                value = line[offset:offset + field_width].strip()
                account_info[field_name] = value
                offset += field_width + 1
            if account_info['Name']:
                match = self._acc_re.match(account_info['Name'])
                if not match:
                    match = self._old_re.match(account_info['Name'])
                if not match:
                    msg = "can not parse account_info name '{0}'\n"
                    sys.stderr.write(msg.format(account_info['Name']))
                else:
                    account_info['Name'] = match.group(1)
            else:
                account_info['Name'] = 'default_project'
            account = MamAccount(acc_id=account_info['Id'],
                                 name=account_info['Name'])
            account.available_credits = account_info['Available']
            account.allocated_credits = account_info['Allocated']
            accounts[account.account_id] = account
        return accounts

